/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package chat.protocol;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.HashMap;
import java.util.Map;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.Push;
import org.objectweb.dream.PushException;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageManager;
import org.objectweb.dream.protocol.ExportIdentifier;
import org.objectweb.dream.protocol.IncomingPush;
import org.objectweb.dream.protocol.OutgoingPush;
import org.objectweb.dream.protocol.channel.ChannelProtocol;
import org.objectweb.dream.protocol.channel.TCPIPProtocol;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;

/**
 * 
 */
public class ChatClientImpl extends AbstractComponent
    implements
      Push,
      IncomingPush,
      ChatClientAttributeController
{

  // ---------------------------------------------------------------------------
  // Attribute field
  // ---------------------------------------------------------------------------

  InetAddress     serverAddress;

  // ---------------------------------------------------------------------------
  // Client interfaces
  // ---------------------------------------------------------------------------

  Push            outPushItf;
  ChannelProtocol lowerProtocolItf;
  MessageManager  messageManagerItf;
  OutgoingPush    outgoingPushItf;

  // ---------------------------------------------------------------------------
  // Constructor
  // ---------------------------------------------------------------------------

  /** Default constructor */
  public ChatClientImpl()
  {
    fcNeedAsyncStart = true;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the Push interface
  // ---------------------------------------------------------------------------

  /**
   * @see Push#push(Message)
   */
  public void push(Message message) throws PushException
  {
    outgoingPushItf.outgoingPush(message);
  }

  // ---------------------------------------------------------------------------
  // Implementation of the IncomingPush interface
  // ---------------------------------------------------------------------------

  /**
   * @see IncomingPush#incomingPush(Message)
   */
  public void incomingPush(Message message) throws PushException
  {
    outPushItf.push(message);
  }

  /**
   * @see IncomingPush#incomingClosed(Object, Exception)
   */
  public void incomingClosed(Object outgoingPush, Exception exception)
  {
    System.err.println("connection closed");
    System.exit(1);
  }

  // ---------------------------------------------------------------------------
  // Implementation of the ChatClientAttributeController interface
  // ---------------------------------------------------------------------------

  /**
   * @see ChatClientAttributeController#getServerHostname()
   */
  public String getServerHostname()
  {
    return serverAddress.getCanonicalHostName();
  }

  /**
   * @see ChatClientAttributeController#setServerHostname(String)
   */
  public void setServerHostname(String serverHostname)
      throws UnknownHostException
  {
    serverAddress = InetAddress.getByName(serverHostname);
  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{ChannelProtocol.LOWER_PROTOCOL_ITF_NAME,
        Push.OUT_PUSH_ITF_NAME, MessageManager.ITF_NAME};
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(ChannelProtocol.LOWER_PROTOCOL_ITF_NAME))
    {
      lowerProtocolItf = (ChannelProtocol) serverItf;
    }
    else if (clientItfName.equals(MessageManager.ITF_NAME))
    {
      messageManagerItf = (MessageManager) serverItf;
    }
    else if (clientItfName.equals(Push.OUT_PUSH_ITF_NAME))
    {
      outPushItf = (Push) serverItf;
    }
  }

  // ---------------------------------------------------------------------------
  // Overridden methods
  // ---------------------------------------------------------------------------

  /**
   * @see AbstractComponent#startFc()
   */
  public void startFc() throws IllegalLifeCycleException
  {
    super.startFc();
    try
    {
      Map<String, Object> info = new HashMap<String, Object>();
      info.put(TCPIPProtocol.ADDRESS, serverAddress);
      info.put(TCPIPProtocol.PORT, new Integer(3000));
      ExportIdentifier identifier = lowerProtocolItf.createExportIdentifier(
          info, ExportIdentifier.EMPTY_EXPORT_IDENTIFIER_ARRAY);
      outgoingPushItf = lowerProtocolItf.bind(identifier, this, null);

    }
    catch (Exception e)
    {
      throw new IllegalLifeCycleException(
          "Unable to bind to exported channel interface : " + e.getMessage());
    }

  }
}
