/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.channel;

import java.util.HashMap;
import java.util.Map;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.Push;
import org.objectweb.dream.PushException;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.protocol.ExportException;
import org.objectweb.dream.protocol.ExportIdentifier;
import org.objectweb.dream.protocol.IncomingPush;
import org.objectweb.dream.protocol.InvalidExportIdentifierException;
import org.objectweb.dream.protocol.Protocol;
import org.objectweb.dream.protocol.channel.ChannelProtocol;
import org.objectweb.dream.protocol.messagePassing.MessagePassingOutgoingPush;
import org.objectweb.dream.protocol.messagePassing.MessagePassingProtocol;
import org.objectweb.dream.util.Error;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;

/**
 * ChannelIn implementation.
 */
public class ChannelInOutImpl extends AbstractComponent
    implements
      Push,
      IncomingPush,
      ChannelInExportId,
      ChannelInAttributeController
{
  private MessagePassingOutgoingPush lowerOutgoingPush;
  private String                     fromChunkName;

  // ---------------------------------------------------------------------------
  // Client interfaces
  // ---------------------------------------------------------------------------

  private MessagePassingProtocol     lowerProtocolItf;
  private DestinationResolver        destinationResolverItf;
  private Push                       outPushItf;

  // ---------------------------------------------------------------------------
  // Constructor
  // ---------------------------------------------------------------------------

  /**
   * Default constructor.
   */
  public ChannelInOutImpl()
  {
    fcNeedAsyncStart = true;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the Push interface
  // ---------------------------------------------------------------------------

  /**
   * @see Push#push(Message)
   */
  public void push(Message message) throws PushException
  {
    try
    {
      lowerOutgoingPush.outgoingPush(message, destinationResolverItf
          .resolveDestination(message));
    }
    catch (InvalidExportIdentifierException e)
    {
      throw new PushException("Invalid destination for message", e);
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of the IncomingPush interface
  // ---------------------------------------------------------------------------

  /**
   * @see IncomingPush#incomingPush(Message)
   */
  public void incomingPush(Message message) throws PushException
  {
    outPushItf.push(message);
  }

  /**
   * @see IncomingPush#incomingClosed(Object, Exception)
   */
  public void incomingClosed(Object outgoingPush, Exception exception)
  {
    // do nothing.
  }

  // ---------------------------------------------------------------------------
  // Implementation of the ChannelInExportId interface
  // ---------------------------------------------------------------------------

  /**
   * @see ChannelInExportId#getExportIdentifier()
   */
  public ExportIdentifier getExportIdentifier()
  {
    return lowerOutgoingPush.getLocalExportIdentifier();
  }

  // ---------------------------------------------------------------------------
  // Overridden methods
  // ---------------------------------------------------------------------------

  /**
   * @see AbstractComponent#startFc()
   */
  public void startFc() throws IllegalLifeCycleException
  {
    super.startFc();
    try
    {
      Map<String, Object> hints = new HashMap<String, Object>();
      hints.put(MessagePassingProtocol.FROM_CHUNK_NAME, fromChunkName);
      lowerOutgoingPush = lowerProtocolItf.export((IncomingPush) weaveableC
          .getFcInterface(INCOMING_IN_PUSH_ITF_NAME), hints);
    }
    catch (ExportException e)
    {
      throw new IllegalLifeCycleException("Unable to export access point : "
          + e.getMessage());
    }
    catch (NoSuchInterfaceException e)
    {
      Error.bug(logger, e);
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of the AttributeController interface
  // ---------------------------------------------------------------------------

  /**
   * @see ChannelInAttributeController#getFromChunkName()
   */
  public String getFromChunkName()
  {
    return fromChunkName;
  }

  /**
   * @see ChannelInAttributeController#setFromChunkName(String)
   */
  public void setFromChunkName(String fromChunkName)
  {
    if (fromChunkName != null && fromChunkName.equals("null"))
    {
      this.fromChunkName = null;
    }
    else
    {
      this.fromChunkName = fromChunkName;
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{Protocol.LOWER_PROTOCOL_ITF_NAME,
        DestinationResolver.ITF_NAME, Push.OUT_PUSH_ITF_NAME};
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public synchronized void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(ChannelProtocol.LOWER_PROTOCOL_ITF_NAME))
    {
      lowerProtocolItf = (MessagePassingProtocol) serverItf;
    }
    else if (clientItfName.equals(Push.OUT_PUSH_ITF_NAME))
    {
      outPushItf = (Push) serverItf;
    }
    else if (clientItfName.equals(DestinationResolver.ITF_NAME))
    {
      destinationResolverItf = (DestinationResolver) serverItf;
    }
  }
}
