/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): Vivien Quema
 */

package org.objectweb.dream.protocol.messagePassing;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

import org.objectweb.dream.message.AbstractChunk;
import org.objectweb.dream.protocol.ExportIdentifier;

/**
 * Chunk for fragmented messages.
 */
public class FragmentationChunk extends AbstractChunk<FragmentationChunk>
{

  private static final long  serialVersionUID = 4049918246973091893L;

  /** The default name for chunks of this type */
  public static final String DEFAULT_NAME     = "fragmentationChunk";

  private ExportIdentifier   framentFrom;
  private int                fragmentationId  = 0;
  private int                fragmentId       = 0;
  private int                numFrags         = 0;

  // ---------------------------------------------------------------------------
  // Chunk methods
  // ---------------------------------------------------------------------------

  /**
   * Returns the identifier of the sender of the fragment.
   * 
   * @return the identifier of the sender of the fragment.
   */
  public ExportIdentifier getFramentFrom()
  {
    return framentFrom;
  }

  /**
   * Sets the identifier of the sender of the fragment.
   * 
   * @param framentFrom the identifier of the sender of the fragment.
   */
  public void setFramentFrom(ExportIdentifier framentFrom)
  {
    this.framentFrom = framentFrom;
  }

  /**
   * Returns the id of this fragment.
   * 
   * @return the id of this fragment.
   */
  public final int getFragmentId()
  {
    return fragmentId;
  }

  /**
   * Sets the fragment's id.
   * 
   * @param fragId the id to set.
   */
  public final void setFragmentId(int fragId)
  {
    this.fragmentId = fragId;
  }

  /**
   * Returns the fragmentation id.
   * 
   * @return the fragmentation id.
   */
  public final int getFragmentationId()
  {
    return fragmentationId;
  }

  /**
   * Sets the fragmentation id.
   * 
   * @param id the id to set.
   */
  public final void setFragmentationId(int id)
  {
    this.fragmentationId = id;
  }

  /**
   * Returns the number of fragments that make the message this fragment is part
   * of.
   * 
   * @return the number of fragments.
   */
  public final int getNumFrags()
  {
    return numFrags;
  }

  /**
   * Sets the number of fragments that make the message this fragment is part
   * of.
   * 
   * @param numFrags the number of fragments to set.
   */
  public final void setNumFrags(int numFrags)
  {
    this.numFrags = numFrags;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the Chunk interface
  // ---------------------------------------------------------------------------

  /**
   * @see AbstractChunk#transfertStateTo(AbstractChunk)
   */
  public void transfertStateTo(FragmentationChunk newInstance)
  {
    newInstance.setFragmentationId(getFragmentationId());
    newInstance.setFragmentId(getFragmentId());
    newInstance.setNumFrags(getNumFrags());
  }

  /**
   * @see AbstractChunk#newChunk()
   */
  public FragmentationChunk newChunk()
  {
    return new FragmentationChunk();
  }

  /**
   * @see org.objectweb.dream.pool.Recyclable#recycle()
   */
  public void recycle()
  {
    fragmentationId = 0;
    fragmentId = 0;
    numFrags = 0;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the Externalizable method
  // ---------------------------------------------------------------------------

  /**
   * @see java.io.Externalizable#writeExternal(ObjectOutput)
   */
  public void writeExternal(ObjectOutput out) throws IOException
  {
    out.writeInt(fragmentationId);
    out.writeInt(fragmentId);
    out.writeInt(numFrags);
  }

  /**
   * @see java.io.Externalizable#readExternal(ObjectInput)
   */
  public void readExternal(ObjectInput in) throws IOException,
      ClassNotFoundException
  {
    fragmentationId = in.readInt();
    fragmentId = in.readInt();
    numFrags = in.readInt();
  }

  // ---------------------------------------------------------------------------
  // Utility methods
  // ---------------------------------------------------------------------------

  /**
   * @see Object#toString()
   */
  public String toString()
  {
    return "[FRAG: id=" + fragmentationId + ", frag_id=" + fragmentId
        + ", num_frags=" + numFrags + ']';
  }

  /**
   * If the given <code>chunkName</code> is equal to the default name returns
   * {@link #DEFAULT_NAME }, otherwise returns the given string.
   * 
   * @param chunkName the chunk name to filter.
   * @return {@link #DEFAULT_NAME }or the given <code>chunkName</code>.
   */
  public static String filterChunkName(String chunkName)
  {
    if (DEFAULT_NAME == chunkName || DEFAULT_NAME.equals(chunkName))
    {
      return DEFAULT_NAME;
    }
    return chunkName;
  }

}