/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.protocol.messagePassing;

import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.IOPushException;
import org.objectweb.dream.PushException;
import org.objectweb.dream.message.ChunkFactoryReference;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageManager;
import org.objectweb.dream.protocol.BindException;
import org.objectweb.dream.protocol.ExportException;
import org.objectweb.dream.protocol.ExportIdentifier;
import org.objectweb.dream.protocol.ExportIdentifierChunk;
import org.objectweb.dream.protocol.IncomingPush;
import org.objectweb.dream.protocol.InvalidExportIdentifierException;
import org.objectweb.dream.protocol.OutgoingPush;
import org.objectweb.dream.protocol.channel.ChannelFactory;
import org.objectweb.dream.protocol.channel.ChannelProtocol;
import org.objectweb.dream.protocol.channel.CloseChunk;
import org.objectweb.dream.util.Error;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Implementation of a message passing protocol using an underlying
 * {@link ChannelProtocol}. This protocol manage a cache of lower channel
 * sessions.
 */
public class MessagePassingOverChannelImpl extends AbstractComponent
    implements
      MessagePassingProtocol,
      MessagePassingOverChannelAttributeController
{

  /**
   * Chunk name of {@link ExportIdentifierChunk} send in first message of newly
   * opened channel session.
   */
  public static final String                           ID_ANNOUNCEMENT_CHUNK_NAME = "id-announcement";
  private static final boolean                         DEFENSIVE_CHECKS           = true;

  private ChunkFactoryReference<ExportIdentifierChunk> exportIdChunkFactory;
  private ChunkFactoryReference<CloseChunk>            closeChunkFactory;

  // ---------------------------------------------------------------------------
  // Attribute fields
  // ---------------------------------------------------------------------------

  private int                                          nbMaxSession;

  // ---------------------------------------------------------------------------
  // Client interfaces
  // ---------------------------------------------------------------------------

  private MessageManager                               messageManagerItf;
  private ChannelProtocol                              lowerProtocolItf;

  // ---------------------------------------------------------------------------
  // Implementation of the UDPProtocol interface
  // ---------------------------------------------------------------------------

  /**
   * @see MessagePassingProtocol#export(IncomingPush, Map)
   */
  public MessagePassingOutgoingPush export(IncomingPush incomingPushItf,
      Map<String, Object> hints) throws ExportException
  {
    Object o = (hints == null) ? null : hints.get(FROM_CHUNK_NAME);
    if (o != null && !(o instanceof String))
    {
      throw new ExportException(
          "Invalid \'from-chunk-name\' hint value, must be a String");
    }
    SessionManager sessionManager = new SessionManager(incomingPushItf,
        (String) o);
    if (incomingPushItf != null)
    {
      ExportIdentifier id = lowerProtocolItf.export(sessionManager, hints);
      sessionManager.localExportIdentifier = id;
    }
    return sessionManager;
  }

  /**
   * @see org.objectweb.dream.protocol.Protocol#createExportIdentifier(Map,
   *      ExportIdentifier[])
   */
  public ExportIdentifier createExportIdentifier(Map<String, Object> info,
      ExportIdentifier[] next) throws InvalidExportIdentifierException
  {
    return lowerProtocolItf.createExportIdentifier(info, next);
  }

  // ---------------------------------------------------------------------------
  // Utility methods
  // ---------------------------------------------------------------------------

  private ExportIdentifierChunk newExportIdentifierChunk()
  {
    if (exportIdChunkFactory == null)
    {
      exportIdChunkFactory = messageManagerItf
          .getChunkFactory(ExportIdentifierChunk.class);
    }
    return messageManagerItf.createChunk(exportIdChunkFactory);
  }

  private CloseChunk newCloseChunk()
  {
    if (closeChunkFactory == null)
    {
      closeChunkFactory = messageManagerItf.getChunkFactory(CloseChunk.class);
    }
    return messageManagerItf.createChunk(closeChunkFactory);
  }

  // ---------------------------------------------------------------------------
  // Inner Class
  // ---------------------------------------------------------------------------

  private class SessionManager
      implements
        MessagePassingOutgoingPush,
        ChannelFactory
  {
    private CacheEntry                  reusableKey      = new CacheEntry();
    private Map<CacheEntry, CacheEntry> cache            = new HashMap<CacheEntry, CacheEntry>();
    private Set<Session>                anonymousSession = new HashSet<Session>();
    private CacheEntry                  lruHead          = null;
    private CacheEntry                  lruTail          = null;
    private boolean                     closed           = false;
    private String                      fromChunkName;

    private ExportIdentifier            localExportIdentifier;
    private IncomingPush                upperIncomingPushItf;

    SessionManager(IncomingPush upperIncomingPushItf, String fromChunkName)
    {
      this.upperIncomingPushItf = upperIncomingPushItf;
      this.fromChunkName = fromChunkName;
    }

    // -------------------------------------------------------------------------
    // Implementation of the ChannelFactory interface
    // -------------------------------------------------------------------------

    /**
     * @see ChannelFactory#instantiate(OutgoingPush)
     */
    public IncomingPush instantiate(OutgoingPush toClientPush)
        throws BindException
    {
      logger.log(BasicLevel.DEBUG, "New incoming connection");
      synchronized (cache)
      {
        if (closed)
        {
          throw new BindException("Access point has been closed");
        }
        Session session = new Session(toClientPush);
        anonymousSession.add(session);

        if (cache.size() >= nbMaxSession)
        {
          logger.log(BasicLevel.DEBUG, "Cache is full");
          // remove last socket in LRU
          if ((lruTail == null) && (cache.size() >= nbMaxSession))
          {
            // every sessions are in use
            try
            {
              // wait for 1 second if a session is released
              cache.wait(1000);
            }
            catch (InterruptedException e)
            {
              throw new BindException(
                  "Interrupted while waiting for cache space");
            }
          }
          if (lruTail == null)
          {
            // no session released
            throw new BindException("Session cache is full");
          }
          if (cache.size() >= nbMaxSession)
          {
            Session s = lruTail.session;
            cacheRemove(lruTail);
            try
            {
              s.close();
            }
            catch (IOException e)
            {
              logger.log(BasicLevel.WARN,
                  "An error occurs while closing session at end of LRU", e);
            }
          }
        }

        return session;
      }
    }

    // -------------------------------------------------------------------------
    // Implementation of the MessagePassingOutgoingPush interface
    // -------------------------------------------------------------------------

    /**
     * @see MessagePassingOutgoingPush#getLocalExportIdentifier()
     */
    public ExportIdentifier getLocalExportIdentifier()
    {
      return localExportIdentifier;
    }

    /**
     * @see MessagePassingOutgoingPush#outgoingPush(Message, ExportIdentifier)
     */
    public void outgoingPush(Message message, ExportIdentifier to)
        throws InvalidExportIdentifierException, IOPushException
    {
      if (localExportIdentifier != null && localExportIdentifier.equals(to))
      {
        if (fromChunkName != null)
        {
          ExportIdentifierChunk chunk = newExportIdentifierChunk();
          chunk.setExportIdentifier(localExportIdentifier);
          messageManagerItf.addChunk(message, fromChunkName, chunk);
        }
        try
        {
          upperIncomingPushItf.incomingPush(message);
        }
        catch (PushException e)
        {
          throw new IOPushException(e.getMessage());
        }
        return;
      }
      Session session = null;
      boolean sessionInUse = false;
      // look for a session in the cache
      CacheEntry ce;
      synchronized (cache)
      {
        if (closed)
        {
          throw new IOPushException("Access point has been closed.");
        }
        if (logger.isLoggable(BasicLevel.DEBUG))
        {
          logger.log(BasicLevel.DEBUG, "Try to get Session in cache for " + to);
        }
        reusableKey.exportIdentifier = to;
        ce = cache.get(reusableKey);
        if (ce == null)
        {
          // open new session.
          if (cache.size() >= nbMaxSession)
          {
            logger.log(BasicLevel.DEBUG, "Cache is full");
            // remove last socket in LRU
            while ((lruTail == null) && (cache.size() >= nbMaxSession))
            {
              logger.log(BasicLevel.DEBUG, "Every sessions are in use, waits");
              try
              {
                cache.wait();
              }
              catch (InterruptedException e)
              {
                throw new IOPushException(
                    "Interrupted while waiting for cache space");
              }
            }
            if (cache.size() >= nbMaxSession)
            {
              logger.log(BasicLevel.DEBUG, "Close last session in LRU.");
              Session s = lruTail.session;
              cacheRemove(lruTail);
              try
              {
                s.close();
              }
              catch (IOException e)
              {
                logger.log(BasicLevel.WARN,
                    "An error occurs while closing session at end of LRU", e);
              }
            }
          }

          ce = new CacheEntry();
          ce.exportIdentifier = to;
          ce.inUse = true;
          cache.put(ce, ce);
        }
        else
        {
          if (logger.isLoggable(BasicLevel.DEBUG))
          {
            logger.log(BasicLevel.DEBUG, "Session found in cache for " + to);
          }
          if (DEFENSIVE_CHECKS && !to.equals(ce.exportIdentifier))
          {
            Error.bug(logger);
          }
          synchronized (ce)
          {
            if (!ce.inUse)
            {
              ce.inUse = true;
              lruRemove(ce);
              if (logger.isLoggable(BasicLevel.DEBUG))
              {
                logger.log(BasicLevel.DEBUG, "Session ok for " + to);
              }
              session = ce.session;
            }
            else
            {
              sessionInUse = true;
            }
          }
        }
      }
      if (session == null) // no session available
      {
        if (sessionInUse)
        {
          // session found in cache but is in use.
          synchronized (ce)
          {
            ce.waitingThread++;
            while (ce.inUse)
            {
              try
              {
                ce.wait();
              }
              catch (InterruptedException e)
              {
                throw new IOPushException(
                    "Interrupted while waiting for in use session");
              }
            }
            ce.inUse = true;
            ce.waitingThread--;
            // ce.session may be null if an error occurs during previous use
            // of it.
            if (ce.session != null)
            {
              if (logger.isLoggable(BasicLevel.DEBUG))
              {
                logger.log(BasicLevel.DEBUG, "Session ok for " + to);
              }
              session = ce.session;
            }
          }
        }

        if (session == null)
        {
          // session not found in cache, bind and create new session
          if (logger.isLoggable(BasicLevel.DEBUG))
          {
            logger.log(BasicLevel.DEBUG, "Create new session to " + to);
          }
          try
          {
            ce.session = new Session(to, ce);
            OutgoingPush lowerOutgoingPush;
            if (upperIncomingPushItf == null)
            {
              lowerOutgoingPush = lowerProtocolItf.bind(to, null, null);
            }
            else
            {
              lowerOutgoingPush = lowerProtocolItf.bind(to, ce.session, null);
            }
            ce.session.init(lowerOutgoingPush);
            logger.log(BasicLevel.DEBUG, "New session created.");
            session = ce.session;
          }
          catch (BindException e)
          {
            synchronized (cache)
            {
              Object o = cache.remove(ce);
              if (DEFENSIVE_CHECKS && o != ce)
              {
                Error.bug(logger);
              }
            }
            throw new IOPushException(e);
          }
        }
      }

      // push outgoing message
      IOPushException exception = null;
      try
      {
        session.outgoingPush.outgoingPush(message);
      }
      catch (IOPushException e)
      {
        exception = e;
      }

      synchronized (cache)
      {
        // re-put session in cache.
        if (DEFENSIVE_CHECKS && (ce == null))
        {
          Error.bug(logger);
        }
        if (DEFENSIVE_CHECKS && (ce.next != null))
        {
          Error.bug(logger);
        }
        if (DEFENSIVE_CHECKS && (ce.prev != null))
        {
          Error.bug(logger);
        }
        if (DEFENSIVE_CHECKS && (ce.session != session))
        {
          Error.bug(logger);
        }
        if (DEFENSIVE_CHECKS && (ce.inUse != true))
        {
          Error.bug(logger);
        }
        if (logger.isLoggable(BasicLevel.DEBUG))
        {
          logger.log(BasicLevel.DEBUG, "Release socket (error=" + exception
              + ") for " + to);
        }
        synchronized (ce)
        {
          if (exception != null)
          {
            logger.log(BasicLevel.DEBUG,
                "An exception occurs while sending the message.", exception);
            if (ce.replacingSession != null)
            {
              logger.log(BasicLevel.DEBUG, "Session has to be replaced by "
                  + "another one. Close and replace it.");
              try
              {
                session.outgoingPush.outgoingClose(session);
              }
              catch (IOException e)
              {
                // ignore
              }
              ce.session = ce.replacingSession;
              ce.replacingSession = null;
              ce.inUse = false;
              ce.notify();
            }
            else if (ce.waitingThread == 0)
            {
              cache.remove(ce);
              try
              {
                session.outgoingPush.outgoingClose(session);
              }
              catch (IOException e)
              {
                // ignore
              }
            }
            else
            {
              ce.inUse = false;
              ce.session = null;
              ce.notify();
            }
          }
          else
          {
            if (ce.replacingSession != null)
            {
              logger.log(BasicLevel.DEBUG, "Session has to be replaced by "
                  + "another one. Send close request and replace it.");
              ce.session.cacheEntry = null;
              ce.session.closeRequestSent = true;
              Message closeMessage = messageManagerItf.createMessage();
              messageManagerItf.addChunk(closeMessage, CloseChunk.DEFAULT_NAME,
                  newCloseChunk());
              try
              {
                ce.session.outgoingPush.outgoingPush(closeMessage);
              }
              catch (IOPushException e)
              {
                logger.log(BasicLevel.WARN,
                    "exception caught while sending close message.", e);
              }
              ce.session = ce.replacingSession;
              ce.replacingSession = null;
            }
            ce.inUse = false;
            if (ce.waitingThread == 0)
            {
              lruAdd(ce);
            }
            else
            {
              ce.notify();
            }
          }
        }
        cache.notifyAll();
        if (exception != null)
        {
          throw exception;
        }
      }

    }

    /**
     * @see MessagePassingOutgoingPush#outgoingClose(IncomingPush)
     */
    public void outgoingClose(IncomingPush incomingPush) throws IOException
    {
      synchronized (cache)
      {
        logger.log(BasicLevel.DEBUG, "Close local access point.");
        if (localExportIdentifier != null)
        {
          try
          {
            logger.log(BasicLevel.DEBUG, "Unexport lower channel factory.");
            lowerProtocolItf.unexport(localExportIdentifier);
          }
          catch (InvalidExportIdentifierException e)
          {
            throw new ExportException(
                "Unable to unexport previously exported channel factory", e);
          }
        }
        closed = true;

        // close sockets in cache
        Iterator iterator = cache.values().iterator();
        while (iterator.hasNext())
        {
          CacheEntry ce = (CacheEntry) iterator.next();
          try
          {
            if (logger.isLoggable(BasicLevel.DEBUG))
            {
              logger.log(BasicLevel.DEBUG, "Close session to "
                  + ce.session.remoteExportIdentifier);
            }
            ce.session.close();
          }
          catch (IOException e)
          {
            logger.log(BasicLevel.WARN,
                "Exception caught while closing lower session", e);
          }
        }

        // close anonymous session.
        iterator = anonymousSession.iterator();
        while (iterator.hasNext())
        {
          Session session = (Session) iterator.next();
          try
          {
            logger.log(BasicLevel.DEBUG, "Close anonymous session.");
            session.close();
          }
          catch (IOException e)
          {
            logger.log(BasicLevel.WARN,
                "Exception caught while closing lower session", e);
          }
          iterator.remove();
        }
      }
    }

    // -------------------------------------------------------------------------
    // Inner Class
    // -------------------------------------------------------------------------

    protected class CacheEntry
    {

      protected ExportIdentifier exportIdentifier;
      protected Session          session;

      protected CacheEntry       next          = null;
      protected CacheEntry       prev          = null;
      protected boolean          inUse         = false;
      protected int              waitingThread = 0;
      protected Session          replacingSession;

      /**
       * @see Object#equals(Object)
       */
      public boolean equals(Object obj)
      {
        final CacheEntry ce = (CacheEntry) obj;
        return exportIdentifier.equals(ce.exportIdentifier);

      }

      /**
       * @see Object#hashCode()
       */
      public int hashCode()
      {
        return exportIdentifier.hashCode();
      }
    }

    protected class Session implements IncomingPush
    {
      ExportIdentifier remoteExportIdentifier;
      OutgoingPush     outgoingPush;
      CacheEntry       cacheEntry;
      boolean          closeRequestSent = false;

      Session(OutgoingPush outgoingPush)
      {
        this.outgoingPush = outgoingPush;
      }

      Session(ExportIdentifier remoteExportIdentifier, CacheEntry cacheEntry)
      {
        this.cacheEntry = cacheEntry;
        this.remoteExportIdentifier = remoteExportIdentifier;
      }

      void init(OutgoingPush outgoingPush) throws IOPushException
      {
        this.outgoingPush = outgoingPush;
        if (localExportIdentifier != null)
        {
          Message message = messageManagerItf.createMessage();
          ExportIdentifierChunk chunk = newExportIdentifierChunk();
          chunk.setExportIdentifier(localExportIdentifier);
          messageManagerItf
              .addChunk(message, ID_ANNOUNCEMENT_CHUNK_NAME, chunk);
          outgoingPush.outgoingPush(message);
        }
      }

      void close() throws IOException
      {
        outgoingPush.outgoingClose(this);
      }

      // -----------------------------------------------------------------------
      // Implementation of the IncomingPush interface
      // -----------------------------------------------------------------------

      /**
       * @see IncomingPush#incomingClosed(Object, Exception)
       */
      public void incomingClosed(Object outgoingPush, Exception exception)
      {
        synchronized (cache)
        {
          if (logger.isLoggable(BasicLevel.INFO))
          {
            logger.log(BasicLevel.INFO, "Lower session to "
                + remoteExportIdentifier + " closed unexpectedly cause : ",
                exception);
          }
          if (DEFENSIVE_CHECKS && this.outgoingPush != outgoingPush)
          {
            Error.bug(logger);
          }
          if (DEFENSIVE_CHECKS && closeRequestSent && cacheEntry != null)
          {
            Error.bug(logger);
          }
          if (cacheEntry != null)
          {
            cacheRemove(cacheEntry);
          }
          else
          {
            boolean removed = anonymousSession.remove(this);
            if (DEFENSIVE_CHECKS && !closeRequestSent && !removed)
            {
              Error.bug(logger);
            }
          }
        }
      }

      /**
       * @see IncomingPush#incomingPush(Message)
       */
      public void incomingPush(Message message) throws PushException
      {
        synchronized (cache)
        {
          if (closed)
          {
            throw new IOPushException("Access point has been closed.");
          }
          ExportIdentifierChunk remoteIdChunk = (ExportIdentifierChunk) messageManagerItf
              .getChunk(message, ID_ANNOUNCEMENT_CHUNK_NAME);
          if (remoteIdChunk != null)
          {
            if (DEFENSIVE_CHECKS && cacheEntry != null)
            {
              Error.bug(logger);
            }
            remoteExportIdentifier = remoteIdChunk.getExportIdentifier();
            messageManagerItf.removeChunk(message, ID_ANNOUNCEMENT_CHUNK_NAME);
            messageManagerItf.deleteChunk(remoteIdChunk);
            if (logger.isLoggable(BasicLevel.DEBUG))
            {
              logger.log(BasicLevel.DEBUG, "Received remote identifier : "
                  + remoteExportIdentifier);
            }
            reusableKey.exportIdentifier = remoteExportIdentifier;
            CacheEntry ce = cache.get(reusableKey);
            if (ce == null)
            {
              logger.log(BasicLevel.DEBUG, "New remote destination");
              boolean removed = anonymousSession.remove(this);
              if (DEFENSIVE_CHECKS && !removed)
              {
                Error.bug(logger);
              }
              // remote destination was unknown
              ce = new CacheEntry();
              ce.exportIdentifier = remoteExportIdentifier;
              ce.session = this;
              cacheEntry = ce;
              cache.put(ce, ce);
              lruAdd(cacheEntry);
            }
            else
            {
              if (DEFENSIVE_CHECKS && (ce == null))
              {
                Error.bug(logger);
              }
              if (DEFENSIVE_CHECKS && (ce.next != null))
              {
                Error.bug(logger);
              }
              if (DEFENSIVE_CHECKS && (ce.prev != null))
              {
                Error.bug(logger);
              }
              logger.log(BasicLevel.DEBUG, "Remote destination already exist");
              // remote destination already exist
              if (localExportIdentifier.hashCode() < remoteExportIdentifier
                  .hashCode())
              {
                logger.log(BasicLevel.DEBUG,
                    "Close session in cache and replace it by this one");
                boolean removed = anonymousSession.remove(this);
                if (DEFENSIVE_CHECKS && !removed)
                {
                  Error.bug(logger);
                }
                // close the session already in cache :
                // first wait until it is free
                // then if the session is still opened, send a close message
                // finally replace by the new one in the cache entry
                synchronized (ce)
                {
                  if (ce.inUse)
                  {
                    logger.log(BasicLevel.DEBUG, "Session in cache is in use,"
                        + " will be closed and replaced later");
                    ce.replacingSession = this;
                  }
                  else
                  {
                    logger.log(BasicLevel.DEBUG, "Send close request.");
                    if (DEFENSIVE_CHECKS && (ce.session == null))
                    {
                      Error.bug(logger);
                    }
                    ce.session.cacheEntry = null;
                    ce.session.closeRequestSent = true;
                    Message closeMessage = messageManagerItf.createMessage();
                    messageManagerItf.addChunk(closeMessage,
                        CloseChunk.DEFAULT_NAME, newCloseChunk());
                    try
                    {
                      ce.session.outgoingPush.outgoingPush(closeMessage);
                    }
                    catch (IOPushException e)
                    {
                      logger.log(BasicLevel.WARN,
                          "exception caught while sending close message.", e);
                    }
                    ce.session = this;
                  }
                  cacheEntry = ce;
                }
              }
            }
            messageManagerItf.deleteMessage(message);
            return;
          }
          CloseChunk closeChunk = (CloseChunk) messageManagerItf.getChunk(
              message, CloseChunk.DEFAULT_NAME);
          if (closeChunk != null)
          {
            logger.log(BasicLevel.DEBUG, "Received a close session request");
            // close session request
            if (DEFENSIVE_CHECKS && cacheEntry != null)
            {
              Error.bug(logger);
            }
            boolean removed = anonymousSession.remove(this);
            if (DEFENSIVE_CHECKS && !removed)
            {
              Error.bug(logger);
            }
            try
            {
              outgoingPush.outgoingClose(this);
            }
            catch (IOException e)
            {
              logger.log(BasicLevel.WARN,
                  "exception caught while closing lower session.", e);
            }
            messageManagerItf.deleteMessage(message);
            return;
          }
          if (cacheEntry != null)
          {
            synchronized (cacheEntry)
            {
              if (!cacheEntry.inUse)
              {
                // put session at head of LRU
                lruRemove(cacheEntry);
                lruAdd(cacheEntry);
              }
            }
          }
          if (fromChunkName != null && remoteExportIdentifier != null)
          {
            ExportIdentifierChunk chunk = newExportIdentifierChunk();
            chunk.setExportIdentifier(remoteExportIdentifier);
            messageManagerItf.addChunk(message, fromChunkName, chunk);
          }
        }
        upperIncomingPushItf.incomingPush(message);
      }
    }

    // -------------------------------------------------------------------------
    // Utility methods
    // -------------------------------------------------------------------------

    private void cacheRemove(CacheEntry ce)
    {
      if (DEFENSIVE_CHECKS && (ce == null))
      {
        Error.bug(logger);
      }
      Object o = cache.remove(ce);
      synchronized (ce)
      {
        if (DEFENSIVE_CHECKS && (o == null))
        {
          Error.bug(logger);
        }
        if (DEFENSIVE_CHECKS && (o != ce))
        {
          Error.bug(logger);
        }
        if (!ce.inUse)
        {
          lruRemove(ce);
        }
      }
    }

    private void lruRemove(CacheEntry ce)
    {
      if (DEFENSIVE_CHECKS && (ce == null))
      {
        Error.bug(logger);
      }
      if (ce.next == null)
      {
        if (DEFENSIVE_CHECKS && (ce != lruTail))
        {
          Error.bug(logger);
        }
        if (ce.prev == null)
        {
          if (DEFENSIVE_CHECKS && (ce != lruHead))
          {
            Error.bug(logger);
          }
          // lru contains only ce
          lruHead = null;
          lruTail = null;
        }
        else
        {
          lruTail = ce.prev;
          lruTail.next = null;
        }
      }
      else
      {
        // next non null
        if (DEFENSIVE_CHECKS && (ce.next.prev != ce))
        {
          Error.bug(logger);
        }
        ce.next.prev = ce.prev;
        if (ce.prev == null)
        {
          // ce == lruHead
          if (DEFENSIVE_CHECKS && (ce != lruHead))
          {
            Error.bug(logger);
          }
          lruHead = ce.next;
        }
        else
        {
          if (DEFENSIVE_CHECKS && (ce.prev.next != ce))
          {
            Error.bug(logger);
          }
          ce.prev.next = ce.next;
        }
      }
      ce.next = null;
      ce.prev = null;
    }

    private void lruAdd(CacheEntry ce)
    {
      if (DEFENSIVE_CHECKS && (ce == null))
      {
        Error.bug(logger);
      }
      if (lruHead == null)
      {
        // lru empty
        if (DEFENSIVE_CHECKS && (lruTail != null))
        {
          Error.bug(logger);
        }
        lruHead = ce;
        lruTail = ce;
        ce.next = null;
        ce.prev = null;
        return;
      }
      lruHead.prev = ce;
      ce.next = lruHead;
      ce.prev = null;
      lruHead = ce;
    }

  }

  // ---------------------------------------------------------------------------
  // Implementation of the AttributeController interface
  // ---------------------------------------------------------------------------

  /**
   * @see MessagePassingOverChannelAttributeController#getNbMaxSession()
   */
  public int getNbMaxSession()
  {
    return nbMaxSession;
  }

  /**
   * @see MessagePassingOverChannelAttributeController#setNbMaxSession(int)
   */
  public void setNbMaxSession(int nbMaxSession)
  {
    this.nbMaxSession = nbMaxSession;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{MessageManager.ITF_NAME, LOWER_PROTOCOL_ITF_NAME};
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(MessageManager.ITF_NAME))
    {
      messageManagerItf = (MessageManager) serverItf;
    }
    else if (clientItfName.equals(LOWER_PROTOCOL_ITF_NAME))
    {
      lowerProtocolItf = (ChannelProtocol) serverItf;
    }
  }
}
