/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.protocol.messagePassing;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.SocketException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.IOPushException;
import org.objectweb.dream.PushException;
import org.objectweb.dream.control.activity.Util;
import org.objectweb.dream.control.activity.manager.ThreadPoolManager;
import org.objectweb.dream.control.activity.task.AbstractTask;
import org.objectweb.dream.control.activity.task.IllegalTaskException;
import org.objectweb.dream.control.activity.task.thread.ThreadPoolOverflowException;
import org.objectweb.dream.message.ChunkFactoryReference;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageManager;
import org.objectweb.dream.protocol.ExportException;
import org.objectweb.dream.protocol.ExportIdentifier;
import org.objectweb.dream.protocol.ExportIdentifierChunk;
import org.objectweb.dream.protocol.IPExportIdentifier;
import org.objectweb.dream.protocol.IncomingPush;
import org.objectweb.dream.protocol.InvalidExportIdentifierException;
import org.objectweb.dream.serializator.DeSerializator;
import org.objectweb.dream.serializator.Serializator;
import org.objectweb.dream.util.Error;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Implementation of the UDP protocol.
 */
public class UDPProtocolImpl extends AbstractComponent
    implements
      UDPProtocol,
      UDPProtocolAttributeController
{

  private static final boolean                         DEFENSIVE_CHECK = true;

  protected static String                              hostName;

  static
  {
    try
    {
      hostName = InetAddress.getLocalHost().getHostName();
    }
    catch (Exception e)
    {
      e.printStackTrace();
    }
  }

  private Set<Session>                                 sessions        = new HashSet<Session>();
  private Session                                      firstAvailable  = null;
  private Session                                      lastAvailable   = null;

  private ThreadPoolManager                            threadPoolManager;
  private ReaderTask                                   readerTask      = new ReaderTask();

  private ChunkFactoryReference<ExportIdentifierChunk> exportIdChunkFactory;

  // ---------------------------------------------------------------------------
  // Attribute fields
  // ---------------------------------------------------------------------------

  int                                                  port;
  int                                                  recvBufferSize;

  // ---------------------------------------------------------------------------
  // Client interfaces
  // ---------------------------------------------------------------------------

  private MessageManager                               messageManagerItf;
  private Serializator                                 serializatorItf;
  private DeSerializator                               deSerializatorItf;

  // ---------------------------------------------------------------------------
  // Implementation of the UDPProtocol interface
  // ---------------------------------------------------------------------------

  /**
   * @see MessagePassingProtocol#export(IncomingPush, Map)
   */
  public MessagePassingOutgoingPush export(IncomingPush incomingPushItf,
      Map hints) throws ExportException
  {
    Number portN = null;
    Number rangeN = null;
    String fromChunkName = null;
    int port = 0;
    int range = 0;
    if (hints != null)
    {
      Object o = hints.get(PORT);
      if (o != null && !(o instanceof Number))
      {
        throw new ExportException(
            "Invalid \'port\' hint value, must be a Number");
      }
      portN = (Number) o;

      o = hints.get(PORT_RANGE);
      if (o != null && !(o instanceof Number))
      {
        throw new ExportException(
            "Invalid \'range\' hint value, must be a Number");
      }
      rangeN = (Number) o;

      o = hints.get(FROM_CHUNK_NAME);
      if (o != null && !(o instanceof String))
      {
        throw new ExportException(
            "Invalid \'from-chunk-name\' hint value, must be a String");
      }
      fromChunkName = (String) o;
    }

    if (portN != null)
    {
      port = portN.intValue();
      if (rangeN != null)
      {
        range = rangeN.intValue();
      }
    }
    else
    {
      synchronized (this)
      {
        port = this.port;
        if (this.port != 0)
        {
          this.port++;
        }
      }
    }
    return export(incomingPushItf, port, range, fromChunkName);
  }

  /**
   * @see org.objectweb.dream.protocol.Protocol#createExportIdentifier(Map,
   *      ExportIdentifier[])
   */
  public ExportIdentifier createExportIdentifier(Map info,
      ExportIdentifier[] next) throws InvalidExportIdentifierException
  {
    if (next != null && next.length != 0)
    {
      throw new InvalidExportIdentifierException("UDP protocol is a leaf "
          + "in the protocol graph, it cannot have next export identifier");
    }
    Object o;
    if (info == null || (o = info.get(ADDRESS)) == null)
    {
      throw new InvalidExportIdentifierException(
          "Can't find address in info map");
    }
    String hostName;
    if (o instanceof String)
    {
      hostName = (String) o;
    }
    else if (o instanceof InetAddress)
    {
      hostName = ((InetAddress) o).getCanonicalHostName();
    }
    else
    {
      throw new InvalidExportIdentifierException(
          "Invalid address object in info map, must be an InetAddress or"
              + " a String");
    }

    o = info.get(PORT);
    if (o == null)
    {
      throw new InvalidExportIdentifierException("Can't find port in info map");

    }
    if (!(o instanceof Number))
    {
      throw new InvalidExportIdentifierException(
          "Invalid port object in info map, must be a Number");
    }
    int port = ((Number) o).intValue();
    return createExportIdentifier(hostName, port);
  }

  /**
   * @see UDPProtocol#export(IncomingPush, int, int, String)
   */
  public MessagePassingOutgoingPush export(IncomingPush incomingPushItf,
      int port, int range, String fromChunkChunk) throws ExportException
  {
    Exception exception = null;
    for (int p = port; p <= port + range; p++)
    {
      Session session;
      try
      {
        session = new Session(incomingPushItf, p, fromChunkChunk);
      }
      catch (SocketException e)
      {
        exception = e;
        continue;
      }
      initSession(session);
      return session;
    }
    throw new ExportException("Unable to open datagram socket", exception);
  }

  /**
   * @see UDPProtocol#createExportIdentifier(String, int)
   */
  public IPExportIdentifier createExportIdentifier(String hostName, int port)
  {
    return new IPExportIdentifier(hostName, port);
  }

  // ---------------------------------------------------------------------------
  // Utility methods
  // ---------------------------------------------------------------------------

  private ExportIdentifierChunk newExportIdentifierChunk()
  {
    if (exportIdChunkFactory == null)
    {
      exportIdChunkFactory = messageManagerItf
          .getChunkFactory(ExportIdentifierChunk.class);
    }
    return messageManagerItf.createChunk(exportIdChunkFactory);
  }

  // ---------------------------------------------------------------------------
  // Inner Class
  // ---------------------------------------------------------------------------

  protected class Session implements MessagePassingOutgoingPush
  {
    Session            nextAvailable;

    byte[]             recvBuffer = new byte[recvBufferSize];
    boolean            closed     = false;
    IncomingPush       upperIncomingPushItf;
    DatagramSocket     socket;
    IPExportIdentifier localExportId;
    String             fromChunkName;

    Session(IncomingPush itf, int port, String fromChunkName)
        throws SocketException
    {
      upperIncomingPushItf = itf;
      this.fromChunkName = fromChunkName;
      if (port > 0)
      {
        socket = new DatagramSocket(port);
      }
      else
      {
        socket = new DatagramSocket();
      }
      localExportId = new IPExportIdentifier(hostName, socket.getLocalPort());
    }

    void receiveMessage()
    {
      DatagramPacket packet = new DatagramPacket(recvBuffer, recvBuffer.length);
      Message message;
      try
      {
        socket.receive(packet);
        message = deSerializatorItf.deserialize(recvBuffer);
      }
      catch (IOException e)
      {
        logger.log(BasicLevel.INFO,
            "An error occurs while reading message on UCP socket", e);
        return;
      }

      if (fromChunkName != null)
      {
        ExportIdentifierChunk chunk = newExportIdentifierChunk();
        chunk.setExportIdentifier(new IPExportIdentifier(packet.getAddress()
            .getCanonicalHostName(), packet.getPort()));
        messageManagerItf.addChunk(message, fromChunkName, chunk);
      }
      try
      {
        upperIncomingPushItf.incomingPush(message);
      }
      catch (PushException e)
      {
        logger.log(BasicLevel.ERROR,
            "Exception catch while pushing incoming message. delete message.",
            e);
        messageManagerItf.deleteMessage(message);
      }

    }

    // -------------------------------------------------------------------------
    // Implementation of the MessagePassingOutgoingPush interface
    // -------------------------------------------------------------------------

    /**
     * @see MessagePassingOutgoingPush#getLocalExportIdentifier()
     */
    public ExportIdentifier getLocalExportIdentifier()
    {
      return localExportId;
    }

    /**
     * @see MessagePassingOutgoingPush#outgoingPush(Message, ExportIdentifier)
     */
    public void outgoingPush(Message message, ExportIdentifier to)
        throws InvalidExportIdentifierException, IOPushException
    {
      synchronized (this)
      {
        if (closed)
        {
          throw new IOPushException("Session closed");
        }
      }
      if (!(to instanceof IPExportIdentifier))
      {
        throw new InvalidExportIdentifierException("Not a IPExportIdentifier",
            to);
      }
      IPExportIdentifier ipTo = (IPExportIdentifier) to;
      byte[] msg;
      try
      {
        msg = serializatorItf.serialize(message);
        DatagramPacket packet = new DatagramPacket(msg, msg.length, InetAddress
            .getByName(ipTo.getHostName()), ipTo.getPort());
        socket.send(packet);
        messageManagerItf.deleteMessage(message);
      }
      catch (IOException e)
      {
        throw new IOPushException(e);
      }
    }

    /**
     * @see MessagePassingOutgoingPush#outgoingClose(IncomingPush)
     */
    public synchronized void outgoingClose(IncomingPush incomingPush)
        throws IOException
    {
      closed = true;
      socket.close();
    }

  }

  protected class ReaderTask extends AbstractTask
  {

    public ReaderTask()
    {
      super("TCP/IP reader task");
    }

    // -------------------------------------------------------------------------
    // Implementation of the Task interface
    // -------------------------------------------------------------------------

    /**
     * @see org.objectweb.dream.control.activity.task.Task#execute(Object)
     */
    public Object execute(Object hints) throws InterruptedException
    {
      Session session;
      synchronized (sessions)
      {
        session = getAvailable();
        if (session == null)
        {
          logger.log(BasicLevel.ERROR, "no available session, stop executing");
          return STOP_EXECUTING;
        }
        if (session.closed)
        {
          if (!sessions.remove(session))
          {
            Error.bug(logger);
          }
          return STOP_EXECUTING;
        }
      }
      session.receiveMessage();
      synchronized (sessions)
      {
        setAvailable(session);
      }
      return EXECUTE_AGAIN;
    }

    /**
     * @see org.objectweb.dream.control.activity.task.Task#registered(Object)
     */
    public void registered(Object controlItf)
    {
      setThreadPoolManager((ThreadPoolManager) controlItf);
    }
  }

  // ---------------------------------------------------------------------------
  // Utility methods
  // ---------------------------------------------------------------------------

  protected void setAvailable(Session session)
  {
    if (firstAvailable == null)
    {
      if (DEFENSIVE_CHECK && lastAvailable != null)
      {
        Error.bug(logger);
      }
      firstAvailable = session;
      lastAvailable = session;
    }
    else
    {
      if ((DEFENSIVE_CHECK)
          && (session.nextAvailable != null || lastAvailable.nextAvailable != null))
      {
        Error.bug(logger);
      }
      lastAvailable.nextAvailable = session;
      lastAvailable = session;
    }
  }

  protected Session getAvailable()
  {
    if (firstAvailable == null)
    {
      if (DEFENSIVE_CHECK && lastAvailable != null)
      {
        Error.bug(logger);
      }
      return null;
    }
    Session available = firstAvailable;
    firstAvailable = available.nextAvailable;
    if (firstAvailable == null)
    {
      lastAvailable = null;
    }
    available.nextAvailable = null;
    return available;
  }

  protected void initSession(Session session)
  {
    synchronized (sessions)
    {
      sessions.add(session);
      setAvailable(session);
      if (session.upperIncomingPushItf != null && threadPoolManager != null)
      {
        try
        {
          logger.log(BasicLevel.DEBUG, "add a thread in thread pool");
          threadPoolManager.addThread(readerTask);
        }
        catch (ThreadPoolOverflowException e)
        {
          logger.log(BasicLevel.WARN, "Unable to add reader thread", e);
        }
        catch (IllegalTaskException e)
        {
          Error.bug(logger, e);
        }
      }
    }
  }

  protected void setThreadPoolManager(ThreadPoolManager threadPoolManager)
  {
    this.threadPoolManager = threadPoolManager;
    synchronized (sessions)
    {
      Iterator iterator = sessions.iterator();
      while (iterator.hasNext())
      {
        Session session = (Session) iterator.next();
        if (session.upperIncomingPushItf == null)
        {
          continue;
        }
        try
        {
          logger.log(BasicLevel.DEBUG, "add a thread in thread pool");
          threadPoolManager.addThread(readerTask);
        }
        catch (ThreadPoolOverflowException e)
        {
          logger.log(BasicLevel.WARN, "Unable to add reader thread", e);
        }
        catch (IllegalTaskException e)
        {
          Error.bug(logger, e);
        }
      }
    }
  }

  // -------------------------------------------------------------------------
  // Overridden methods
  // -------------------------------------------------------------------------

  /**
   * @see AbstractComponent#beforeFirstStart(Component)
   */
  protected void beforeFirstStart(Component componentItf)
      throws IllegalLifeCycleException
  {
    try
    {
      Map<String, Object> hints = new HashMap<String, Object>();
      hints.put("thread", "pool");
      Util.addTask(componentItf, readerTask, hints);
    }
    catch (Exception e)
    {
      throw new IllegalLifeCycleException("Can't add task");
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of the AttributeController interface
  // ---------------------------------------------------------------------------

  /**
   * @see UDPProtocolAttributeController#getPort()
   */
  public int getPort()
  {
    return port;
  }

  /**
   * @see UDPProtocolAttributeController#setPort(int)
   */
  public void setPort(int port)
  {
    this.port = port;
  }

  /**
   * @see UDPProtocolAttributeController#getRecvBufferSize()
   */
  public int getRecvBufferSize()
  {
    return recvBufferSize;
  }

  /**
   * @see UDPProtocolAttributeController#setRecvBufferSize(int)
   */
  public void setRecvBufferSize(int recvBufferSize)
  {
    this.recvBufferSize = recvBufferSize;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{MessageManager.ITF_NAME, Serializator.ITF_NAME,
        DeSerializator.ITF_NAME};
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(MessageManager.ITF_NAME))
    {
      messageManagerItf = (MessageManager) serverItf;
    }
    else if (clientItfName.equals(Serializator.ITF_NAME))
    {
      serializatorItf = (Serializator) serverItf;
    }
    else if (clientItfName.equals(DeSerializator.ITF_NAME))
    {
      deSerializatorItf = (DeSerializator) serverItf;
    }
  }

}
