/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.protocol.rpc;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.PushException;
import org.objectweb.dream.message.ChunkFactoryReference;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageManager;
import org.objectweb.dream.protocol.BindException;
import org.objectweb.dream.protocol.ExportException;
import org.objectweb.dream.protocol.ExportIdentifier;
import org.objectweb.dream.protocol.IncomingPush;
import org.objectweb.dream.protocol.InvalidExportIdentifierException;
import org.objectweb.dream.protocol.OutgoingPush;
import org.objectweb.dream.protocol.channel.ChannelFactory;
import org.objectweb.dream.protocol.channel.ChannelProtocol;
import org.objectweb.dream.util.Error;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;

public class ProtocolImpl extends AbstractComponent implements RPCProtocol
{

  private final Map<ExportIdentifier, StubSession> stubSessions      = new HashMap<ExportIdentifier, StubSession>();
  private final Map<ExportIdentifier, Skeleton>    exportedSkeletons = new HashMap<ExportIdentifier, Skeleton>();
  private ChunkFactoryReference<StubClassChunk>    stubClassChunkFactory;
  private ChunkFactoryReference<InvocationChunk>   invocationChunkFactory;
  private ChunkFactoryReference<ReplyChunk>        replyChunkFactory;

  // ---------------------------------------------------------------------------
  // Client interfaces
  // ---------------------------------------------------------------------------

  private ChannelProtocol                          lowerLevelProtocolItf;
  private MessageManager                           messageManagerItf;

  // ---------------------------------------------------------------------------
  // Implementation of the RPCProtocol interface
  // ---------------------------------------------------------------------------

  /**
   * @see RPCProtocol#export(Skeleton, Map)
   */
  public ExportIdentifier export(Skeleton skeleton, Map<String, Object> hints)
      throws ExportException
  {
    SkeletonChannel skeletonChannel = new SkeletonChannel(skeleton);
    ExportIdentifier id = lowerLevelProtocolItf.export(skeletonChannel, hints);
    synchronized (this)
    {
      exportedSkeletons.put(id, skeleton);
    }
    return id;
  }

  /**
   * @see RPCProtocol#bind(ExportIdentifier, Map)
   */
  public Stub bind(ExportIdentifier exportId, Map<String, Object> hints)
      throws InvalidExportIdentifierException, BindException
  {
    Skeleton skeleton;
    StubSession stubSession = null;
    boolean newStubSession = false;
    synchronized (this)
    {
      skeleton = exportedSkeletons.get(exportId);
      if (skeleton == null)
      {
        stubSession = stubSessions.get(exportId);
        if (stubSession == null)
        {
          stubSession = new StubSession();
          stubSessions.put(exportId, stubSession);
          newStubSession = true;
        }
      }
    }
    if (skeleton != null)
    {
      // bind to a locally exported interface
      Stub stub;
      try
      {
        stub = (Stub) skeleton.getStubClass().newInstance();
      }
      catch (Exception e)
      {
        throw new BindException("Unable to instantiate stub + ", exportId, e);
      }
      stub.setInvoker(skeleton);
      return stub;
    }
    if (newStubSession)
    {
      // new non local binding;
      OutgoingPush lowerPush;
      try
      {
        lowerPush = lowerLevelProtocolItf.bind(exportId, stubSession, hints);
      }
      catch (BindException e)
      {
        synchronized (this)
        {
          stubSessions.remove(exportId);
        }
        synchronized (stubSession)
        {
          stubSession.exception = e;
          stubSession.notifyAll();
        }
        throw e;
      }
      catch (InvalidExportIdentifierException e)
      {
        synchronized (this)
        {
          stubSessions.remove(exportId);
        }
        synchronized (stubSession)
        {
          stubSession.exception = e;
          stubSession.notifyAll();
        }
        throw e;
      }
      stubSession.lowerOutgoingPush = lowerPush;
    }
    try
    {
      return stubSession.getStub();
    }
    catch (Exception e)
    {
      // TODO see what to clean...
      throw new BindException("An exception occurs while instantiating stub",
          exportId, e);
    }
  }

  /**
   * @see RPCProtocol#createExportIdentifier(Map, ExportIdentifier[])
   */
  public ExportIdentifier createExportIdentifier(Map<String, Object> info,
      ExportIdentifier[] next) throws InvalidExportIdentifierException
  {
    // TODO Auto-generated method stub
    return null;
  }

  // ---------------------------------------------------------------------------
  // Utility methods
  // ---------------------------------------------------------------------------

  StubClassChunk newStubClassChunk()
  {
    if (stubClassChunkFactory == null)
    {
      stubClassChunkFactory = messageManagerItf
          .getChunkFactory(StubClassChunk.class);
    }
    return messageManagerItf.createChunk(stubClassChunkFactory);
  }

  InvocationChunk newInvocationChunk()
  {
    if (invocationChunkFactory == null)
    {
      invocationChunkFactory = messageManagerItf
          .getChunkFactory(InvocationChunk.class);
    }
    return messageManagerItf.createChunk(invocationChunkFactory);
  }

  ReplyChunk newReplyChunk()
  {
    if (replyChunkFactory == null)
    {
      replyChunkFactory = messageManagerItf.getChunkFactory(ReplyChunk.class);
    }
    return messageManagerItf.createChunk(replyChunkFactory);
  }

  // ---------------------------------------------------------------------------
  // Inner classes
  // ---------------------------------------------------------------------------

  protected class SkeletonChannel implements ChannelFactory
  {

    Skeleton skeletonItf;

    public SkeletonChannel(Skeleton itf)
    {
      skeletonItf = itf;
    }

    // -------------------------------------------------------------------------
    // Implementation of the Channel interface
    // -------------------------------------------------------------------------

    /**
     * @see ChannelFactory#instantiate(OutgoingPush)
     */
    public IncomingPush instantiate(OutgoingPush toClientPush)
        throws BindException
    {
      Message message = messageManagerItf.createMessage();
      StubClassChunk stubClassChunk = newStubClassChunk();
      stubClassChunk.setStubClass(skeletonItf.getStubClass());
      messageManagerItf.addChunk(message, StubClassChunk.DEFAULT_NAME,
          stubClassChunk);
      try
      {
        toClientPush.outgoingPush(message);
      }
      catch (PushException e)
      {
        throw new BindException("Unable to send stub class to client session");
      }
      Session session = new Session(toClientPush);
      return session;
    }

    protected class Session implements IncomingPush
    {

      OutgoingPush outgoingPushItf;
      boolean      closed = false;

      protected Session(OutgoingPush itf)
      {
        outgoingPushItf = itf;
      }

      /**
       * @see IncomingPush#incomingPush(Message)
       */
      public void incomingPush(Message message) throws PushException
      {
        InvocationChunk chunk = (InvocationChunk) messageManagerItf
            .removeChunk(message, InvocationChunk.DEFAULT_NAME);
        messageManagerItf.deleteMessage(message);
        ReplyChunk replyChunk;
        try
        {
          replyChunk = skeletonItf.invoke(chunk);
        }
        catch (InvocationException e)
        {
          replyChunk = newReplyChunk();
          replyChunk.setInvocationId(chunk.getInvocationId());
          replyChunk.setThrowable(e);
        }
        messageManagerItf.deleteChunk(chunk);
        if (!closed)
        {
          Message replyMessage = messageManagerItf.createMessage();
          messageManagerItf.addChunk(replyMessage, ReplyChunk.DEFAULT_NAME,
              replyChunk);
          try
          {
            outgoingPushItf.outgoingPush(replyMessage);
          }
          catch (PushException e)
          {
            messageManagerItf.deleteMessage(replyMessage);
            logger.log(BasicLevel.ERROR, "An error occurs while sending reply",
                e);
          }
        }
        else
        {
          logger.log(BasicLevel.ERROR,
              "Session closed while invoking method, cannot send reply");
        }
      }

      /**
       * @see IncomingPush#incomingClosed(Object, Exception)
       */
      public synchronized void incomingClosed(Object outgoingPush,
          Exception exception)
      {
        closed = true;
      }

    }
  }

  protected class StubSession implements Invoker, IncomingPush
  {
    Map<Long, InvocationSemaphore> waitingInvocation = new HashMap<Long, InvocationSemaphore>();
    long                           nextInvocationId  = 0;
    Stub                           stub              = null;
    Exception                      exception         = null;

    OutgoingPush                   lowerOutgoingPush;

    public ReplyChunk invoke(InvocationChunk invocationChunk)
        throws InvocationException
    {
      long id = getNextInvocationId();
      InvocationSemaphore invocation = new InvocationSemaphore();
      synchronized (this)
      {
        waitingInvocation.put(id, invocation);
      }

      invocationChunk.setInvocationId(id);
      Message message = messageManagerItf.createMessage();
      messageManagerItf.addChunk(message, InvocationChunk.DEFAULT_NAME,
          invocationChunk);
      try
      {
        lowerOutgoingPush.outgoingPush(message);
      }
      catch (PushException e)
      {
        throw new InvocationException(
            "An exception occurs while sending invocation request", e);
      }
      synchronized (invocation)
      {
        while (invocation.reply == null)
        {
          try
          {
            invocation.wait();
          }
          catch (InterruptedException e)
          {
            throw new InvocationException(
                "Interrupted while waiting for invocation response");
          }
        }
      }
      return invocation.reply;
    }

    public void incomingPush(Message message) throws PushException
    {
      InvocationSemaphore invocation;
      ReplyChunk chunk;
      synchronized (this)
      {
        if (stub == null)
        {
          StubClassChunk stubClassChunk = (StubClassChunk) messageManagerItf
              .getChunk(message, StubClassChunk.DEFAULT_NAME);
          if (stubClassChunk == null)
          {
            Error.error(
                "First received message does not contains the stub class",
                logger);
          }
          try
          {
            stub = (Stub) stubClassChunk.getStubClass().newInstance();
          }
          catch (Exception e)
          {
            exception = e;
          }
          messageManagerItf.deleteMessage(message);
          stub.setInvoker(this);
          notifyAll();
          return;
        }
        chunk = (ReplyChunk) messageManagerItf.getChunk(message,
            ReplyChunk.DEFAULT_NAME);
        long id = chunk.getInvocationId();
        messageManagerItf.deleteMessage(message);
        invocation = waitingInvocation.remove(id);
      }
      if (invocation == null)
      {
        logger.log(BasicLevel.ERROR, "Unable to find waiting invocation "
            + "for the given invocation identifier.");
      }
      synchronized (invocation)
      {
        invocation.reply = chunk;
        invocation.notify();
      }
    }

    public synchronized void incomingClosed(Object outgoingPush,
        Exception exception)
    {
      Iterator iterator = waitingInvocation.entrySet().iterator();
      while (iterator.hasNext())
      {
        Map.Entry entry = (Entry) iterator.next();
        long id = ((Long) entry.getKey()).longValue();
        InvocationSemaphore invocation = (InvocationSemaphore) entry.getValue();
        ReplyChunk replyChunk = newReplyChunk();
        replyChunk.setInvocationId(id);
        replyChunk
            .setThrowable(new InvocationException(
                "The communication channel has been closed while waiting for reply",
                exception));
        synchronized (invocation)
        {
          invocation.reply = replyChunk;
          invocation.notify();
        }
        iterator.remove();
      }
    }

    protected synchronized Stub getStub() throws Exception
    {
      while (stub == null && exception == null)
      {
        wait();
      }
      if (exception != null)
      {
        throw exception;
      }
      return stub;
    }

    private synchronized long getNextInvocationId()
    {
      return nextInvocationId++;
    }

    protected class InvocationSemaphore
    {
      ReplyChunk reply;
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{MessageManager.ITF_NAME,
        ChannelProtocol.LOWER_PROTOCOL_ITF_NAME};
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    if (clientItfName.equals(MessageManager.ITF_NAME))
    {
      messageManagerItf = (MessageManager) serverItf;
    }
    else if (clientItfName.equals(ChannelProtocol.LOWER_PROTOCOL_ITF_NAME))
    {
      lowerLevelProtocolItf = (ChannelProtocol) serverItf;
    }
    super.bindFc(clientItfName, serverItf);
  }
}
