/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s): Romain Lenglet
 */

package org.objectweb.dream.queue.keyed;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageManager;
import org.objectweb.dream.queue.BufferAttributeController;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;

/**
 * Abstract implementation of a keyed buffer. Keyed buffers allow mapping
 * messages to keys. This class implements the {@link BufferAttributeController},
 * {@link KeyedBufferAdd} and {@link KeyedBufferRemove} interfaces. When a
 * message is added to the buffer, an associated key is specified. This key can
 * later be used to retrieve the message.
 * <p>
 * The <code>add</code> and <code>remove</code> methods are implemented as
 * follows
 * <ul>
 * <li>The <code>add</code> method checks for available space and then calls
 * the <code>doAdd</code> method to add the message to the buffer.</li>
 * <li>The <code>remove</code> method checks for available messages and then
 * calls the <code>doRemove</code> method to remove a message from the buffer.
 * </li>
 * </ul>
 * <p>
 * Buffer developers can inherit this class. They must implement several
 * methods:
 * <ul>
 * <li><code>canAdd(message, key)</code> returns a boolean indicating whether
 * the given message can be added into the buffer.</li>
 * <li><code>doAdd(message, key)</code> adds the given message into the
 * buffer.</li>
 * <li><code>hasAvailableMessage(key)</code> returns a boolean indicating
 * whether there is an available message mapping the given key.</li>
 * <li><code>doRemove(key)</code> removes a message from the buffer.</li>
 * <li>etc.</li>
 * </ul>
 * <p>
 * Moreover developers must use the <code>incrementAvailableMessagesCount</code>
 * and <code>incrementStoredMessagesCount</code> methods to increment
 * indicators on stored messages, available messages and available space.
 * <p>
 * <i>Note: </i> the <code>add</code> and <code>remove</code> methods should
 * not be overridden.
 */
public abstract class AbstractBufferKeyedAddKeyedRemoveImpl
    extends
      AbstractComponent
    implements
      KeyedBufferAdd,
      KeyedBufferRemove,
      BufferAttributeController
{

  // ------------------------------------------------------------------------
  // Constants.
  // ------------------------------------------------------------------------

  /**
   * The names of client interfaces.
   */
  protected static final String[] CLIENT_INTERFACE_NAMES = {MessageManager.ITF_NAME};

  // ------------------------------------------------------------------------
  // Client interfaces.
  // ------------------------------------------------------------------------

  /**
   * Can be used by subclasses to create aggregated messages in {@link
   * KeyedBufferRemove#removeAll()}.
   */
  protected MessageManager        messageManagerItf;

  // ------------------------------------------------------------------------
  // Attributes.
  // ------------------------------------------------------------------------

  /**
   * The maximal capacity of the buffer.
   */
  protected int                   maxCapacity            = 0;

  // ------------------------------------------------------------------------
  // Fields.
  // ------------------------------------------------------------------------

  /**
   * The count of available messages in the buffer. This attribute should only
   * be updated using the {@link #incrementAvailableMessagesCount(int, Object)}
   * method.
   */
  protected int                   availableMessagesCount = 0;

  /**
   * The count of stored messages into the buffer. This attribute should only be
   * updated using the {@link #incrementStoredMessagesCount(int, Object)}
   * method.
   */
  protected int                   storedMessagesCount    = 0;

  /**
   * The lock object.
   */
  protected final Object          lock                   = new Object();

  // ------------------------------------------------------------------------
  // Implementation of the KeyedBufferAdd interface.
  // ------------------------------------------------------------------------

  /**
   * @see KeyedBufferAdd#add(Message, Object)
   */
  public void add(Message message, Object key) throws InterruptedException
  {
    synchronized (lock)
    {
      while (!canAdd(message, key))
      {
        lock.wait();
      }
      doAdd(message, key);
      lock.notifyAll();
    }
  }

  /**
   * @see KeyedBufferAdd#tryAdd(Message, Object)
   */
  public boolean tryAdd(Message message, Object key)
  {
    synchronized (lock)
    {
      if (canAdd(message, key))
      {
        doAdd(message, key);
        lock.notifyAll();
        return true;
      }
      return false;
    }
  }

  // ------------------------------------------------------------------------
  // Implementation of the KeyedBufferRemove interface.
  // ------------------------------------------------------------------------

  /**
   * @see KeyedBufferRemove#remove(Object)
   */
  public Message remove(Object key) throws InterruptedException
  {
    synchronized (lock)
    {
      while (!hasAvailableMessage(key))
      {
        lock.wait();
      }
      final Message m = doRemove(key);
      lock.notifyAll();
      return m;
    }
  }

  /**
   * @see KeyedBufferRemove#tryRemove(Object)
   */
  public Message tryRemove(Object key)
  {
    synchronized (lock)
    {
      if (hasAvailableMessage(key))
      {
        final Message m = doRemove(key);
        lock.notifyAll();
        return m;
      }
      return null;
    }
  }

  /**
   * @see KeyedBufferRemove#get(Object)
   */
  public Message get(Object key) throws InterruptedException
  {
    synchronized (lock)
    {
      while (!hasAvailableMessage(key))
      {
        lock.wait();
      }
      final Message m = doGet(key);
      return m;
    }
  }

  /**
   * @see KeyedBufferRemove#tryGet(Object)
   */
  public Message tryGet(Object key)
  {
    synchronized (lock)
    {
      if (hasAvailableMessage(key))
      {
        final Message m = doGet(key);
        return m;
      }
      return null;
    }
  }

  /**
   * @see KeyedBufferRemove#removeAll()
   */
  public Message removeAll()
  {
    synchronized (lock)
    {
      final Message m = doRemoveAll();
      lock.notifyAll();
      return m;
    }
  }

  // ------------------------------------------------------------------------
  // Methods that must be implemented by concrete sub classes.
  // ------------------------------------------------------------------------

  /**
   * Adds a message to the buffer. This method should not check if there is
   * enough available space provided it has already been done in the
   * {@link #add}method. The {@link #lock lock}has been acquired when this
   * method is called.
   * 
   * @param message the message to be added.
   */
  protected abstract void doAdd(Message message, Object key);

  /**
   * Removes a message from the buffer. This method should not check if there is
   * a message available provided it has already been done in the
   * {@link #remove}method. The {@link #lock lock}has been acquired when this
   * method is called.
   * 
   * @return a message.
   */
  protected abstract Message doRemove(Object key);

  /**
   * Removes all the messages stored into the buffer. The {@link #lock lock}has
   * been acquired when this method is called.
   * 
   * @return an aggregated message.
   */
  protected abstract Message doRemoveAll();

  /**
   * Gets a message from the buffer. This method should not check if there is a
   * message available provided it has already been done in the {@link #get}
   * method. The {@link #lock lock}has been acquired when this method is
   * called.
   * 
   * @return a message.
   */
  protected abstract Message doGet(Object key);

  // ------------------------------------------------------------------------
  // Methods that may be overridden by concrete sub classes.
  // ------------------------------------------------------------------------

  /**
   * Checks whether the given message can be added into the buffer. The
   * {@link #lock lock}has been acquired when this method is called.
   * 
   * @param message the message to be tested.
   * @return <code>true</code> if the message can be added.
   */
  protected boolean canAdd(Message message, Object key)
  {
    if (maxCapacity > 0 && (maxCapacity - storedMessagesCount <= 0))
    {
      return false;
    }
    return true;
  }

  /**
   * Checks whether there is a message available to get or remove.
   * 
   * @return <code>true</code> if there is an available message.
   */
  protected boolean hasAvailableMessage(Object key)
  {
    return (availableMessagesCount > 0);
  }

  // ------------------------------------------------------------------------
  // Utility methods.
  // ------------------------------------------------------------------------

  protected int availableSpace(Object key)
  {
    if (maxCapacity <= 0)
    {
      return Integer.MAX_VALUE;
    }
    return maxCapacity - storedMessagesCount;
  }

  /**
   * Increments the count of available messages. When calling this method, the
   * {@link #lock lock}must have already been acquired.
   * 
   * @param delta the value that must be added to the previous count.
   */
  protected void incrementAvailableMessagesCount(int delta, Object key)
  {
    availableMessagesCount += delta;
    /*
     * It is assumed that the locked has already been acquired.
     */
    lock.notifyAll();
  }

  /**
   * Increments the count of stored messages. When calling this method, the
   * {@link #lock lock}must have already been acquired.
   * 
   * @param delta the value that must be added to previous indicator.
   */
  protected void incrementStoredMessagesCount(int delta, Object key)
  {
    storedMessagesCount += delta;
    /*
     * It is assumed that the locked has already been acquired.
     */
    lock.notifyAll();
  }

  // ------------------------------------------------------------------------
  // Implementation of the BufferAttributeController interface.
  // ------------------------------------------------------------------------

  /**
   * @see BufferAttributeController#getMaxCapacity()
   */
  public int getMaxCapacity()
  {
    return maxCapacity;
  }

  /**
   * @see BufferAttributeController#setMaxCapacity(int)
   */
  public void setMaxCapacity(int maxCapacity)
  {
    synchronized (lock)
    {
      int previousCapacity = this.maxCapacity;
      this.maxCapacity = maxCapacity;
      if (maxCapacity > previousCapacity)
      {
        lock.notifyAll();
      }
    }
  }

  /**
   * @see BufferAttributeController#getCurrentSize()
   */
  public int getCurrentSize()
  {
    return storedMessagesCount;
  }

  // ------------------------------------------------------------------------
  // Implementation of the BindingController interface.
  // ------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public synchronized void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(MessageManager.ITF_NAME))
    {
      messageManagerItf = (MessageManager) serverItf;
    }
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return CLIENT_INTERFACE_NAMES;
  }

}
