/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema 
 * Contributor(s): Romain Lenglet
 */

package org.objectweb.dream.queue.keyed;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.PullException;
import org.objectweb.dream.message.Message;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;

/**
 * An outgoing handler that implements the {@link KeyedPull}interface. It
 * {@link #pull(Object) returns}messages removed from a
 * {@link KeyedBufferRemove buffer}bound to it. This component uses
 * {@link KeyedBufferRemove#remove(Object) blocking calls for removing}messages
 * from the buffer.
 */
public class KeyedOutgoingHandlerBlockingImpl extends AbstractComponent
    implements
      KeyedPull
{

  // ---------------------------------------------------------------------------
  // Constants.
  // ---------------------------------------------------------------------------

  /**
   * The names of client interfaces.
   */
  protected static final String[] CLIENT_INTERFACE_NAMES = {KeyedBufferRemove.ITF_NAME};

  // ---------------------------------------------------------------------------
  // Client interfaces.
  // ---------------------------------------------------------------------------

  /**
   * The buffer from which pulled messages are removed.
   */
  protected KeyedBufferRemove     keyedBufferRemoveItf;

  // ---------------------------------------------------------------------------
  // Implementation of the KeyedPull interface.
  // ---------------------------------------------------------------------------

  /**
   * @see KeyedPull#pull(Object)
   */
  public Message pull(Object key) throws PullException
  {
    try
    {
      return keyedBufferRemoveItf.remove(key);
    }
    catch (InterruptedException e)
    {
      throw new PullException(
          "interrupted while removing message from the buffer", e);
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface.
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return CLIENT_INTERFACE_NAMES;
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public synchronized void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(KeyedBufferRemove.ITF_NAME))
    {
      keyedBufferRemoveItf = (KeyedBufferRemove) serverItf;
    }
  }
}