/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s): Romain Lenglet
 */

package org.objectweb.dream.router;

import java.util.Hashtable;
import java.util.Map;

import org.objectweb.dream.Push;
import org.objectweb.dream.PushException;
import org.objectweb.dream.message.Message;

/**
 * Abstract implementation of a key-based Push/Push router. A key-based
 * Push/Push router associates each of its outputs (except the default one) to a
 * key. These mappings (key,outputs) are stored in the
 * <code>outPushRouteTable</code> Map. The key corresponding to an incoming
 * message is retrieved using the abstract
 * {@link AbstractRouterKeyBasedImpl#getOutputKey(Message)}method. <br>
 * The <code>outPushRouteTable</code> Map is intended to be built by the
 * {@link AbstractRouterKeyBasedImpl#initRouteTable() }method. To avoid
 * unnecessary building of the <code>outPushRouteTable</code> Map, the
 * {@link AbstractRouterKeyBasedImpl#initRouteTable() }should test the
 * <code>initialized</code> boolean which is set to <code>false</code> each
 * time the router is bound to a new output.
 */
public abstract class AbstractRouterKeyBasedImpl extends AbstractRouterImpl
{

  /**
   * Route table that associates keys with <code>Push</code> output
   * interfaces.
   */
  protected Map outPushRouteTable;

  /**
   * Constructor.
   */
  public AbstractRouterKeyBasedImpl()
  {
    super();
    outPushRouteTable = new Hashtable();
  }

  /**
   * @see org.objectweb.dream.router.AbstractRouterImpl#getOutput(Message)
   */
  protected Push getOutput(Message message) throws PushException
  {
    initRouteTable();
    Object key = getOutputKey(message);
    if (key == null)
    {
      return null;
    }
    return (Push) outPushRouteTable.get(key);
  }

  /**
   * Returns the key associated with the output on which the specified message
   * must be routed.
   * 
   * @param message the message that must be routed.
   * @return the key associated with the output on which the specified message
   *         must be routed.
   */
  protected abstract Object getOutputKey(Message message);

  /**
   * This method initializes the <code>outPushRouteTable</code> route table.
   * It is called by the <code>getOutput</code> method. To avoid unnecessary
   * route table, this method should test the <code>initialized</code> boolean
   * which is set to <code>false</code> each time the router is bound to a new
   * output.
   */
  protected void initRouteTable() throws PushException
  {
  }

}